# Used to system info and other data
# Variables are handled in varproc.py

# THIS FILE SHOULD NOT IMPORT DPL MODULES AS A CIRCULAR IMPORT WILL HAPPEN
# ALMOST EVERY FILE IN lib.core IMPORTS THIS

import os, sys
import platform
from . import constants

ARGV = sys.argv
ARGC = len(ARGV)

INC_EXT = {
    "fn",
    "method",
    "for",
    "loop",
    "while",
    "if",
    "if_else",
    "thread",
    "ismain",
    "isntmain",
    "expect",
    "expect-then",
    "body",
    "template",
    "from_template",
}

INC = {"if-then": 2, "thread": 1, "expect-then": 2}

DEC = {"end", "then", "else"}

CHARS = {
    "\\\\": "\\[escape]",
    "\\n": "\n",
    "\\t": "\t",
    "\\s": " ",
    "\\v": "\v",
    "\\f": "\f",
    "\\r": "\r",
    "\\a": "\a",
    "\\0": "\0",
    "\\[win_nl]": "\r\n",
    "\\[posix_nl]": "\n",
    "\\[null]": "\0",
    "\\[alert]": "\a",
    "\\[escape]": "\\",
}

WARNINGS = True
DEAD_CODE_OPT = True

VERSION_TRIPLE = (1, 4, 5)


def isCompat(version, VERSION=VERSION_TRIPLE):
    major, minor, patch = version
    if major != VERSION[0]:
        return False
    elif minor != VERSION[1]:
        return False
    if VERSION[2] is None or patch is None:
        return True
    if patch >= VERSION[2]:
        return True
    else:
        return False


def getDiff(version_triple, VERSION=VERSION_TRIPLE):
    major, minor, patch = version_triple
    if major < VERSION[0]:
        return "Script is outdated!"
    elif major > VERSION[0]:
        return "Interpreter is outdated!"
    elif minor < VERSION[1]:
        return "Script is outdated!"
    elif minor > VERSION[1]:
        return "Interpreter is outdated!"
    if VERSION[2] is None or patch is None:
        return 0
    if patch < VERSION[2]:
        return "Script is outdated!"
    else:
        return 0


def isLater(version_triple, VERSION=VERSION_TRIPLE):
    major, minor, patch = version_triple
    b_major, b_minor, b_patch = VERSION

    b_patch = b_patch or 0

    if major >= b_major:
        if minor >= b_minor:
            if b_patch is None or patch is None:
                return True
            if patch >= b_patch:
                return True
            else:
                return False
        else:
            return False
    else:
        return False


class Version:
    def __init__(self, major, minor, patch=None):
        self.ver = (major, minor, patch)

    def isCompat(self, version_triple):
        if isinstance(version_triple, tuple):
            return isCompat(version_triple, self.ver)
        elif isinstance(version_triple, (Version, VersionSpec)):
            return isCompat(version_triple.ver, self.ver)

    def isLater(self, version_triple):
        if isinstance(version_triple, tuple):
            return isLater(version_triple, self.ver)
        elif isinstance(version_triple, (Version, VersionSpec)):
            return isLater(version_triple.ver, self.ver)

    def getDiff(self, version_triple, VERSION=None):
        VERSION = VERSION
        major, minor, patch = version_triple
        if major < VERSION[0]:
            return "Script is outdated!"
        elif major > VERSION[0]:
            return "Interpreter is outdated!"
        elif minor < VERSION[1]:
            return "Script is outdated!"
        elif minor > VERSION[1]:
            return "Interpreter is outdated!"
        if VERSION[2] is None or patch is None:
            return 0
        if patch < VERSION[2]:
            return "Script is outdated!"
        else:
            return 0

    def __repr__(self):
        if self.ver[2] is None:
            return ".".join(map(str, self.ver[:2])) + ".x"
        return ".".join(map(str, self.ver))


class VersionSpec(Version):
    def __init__(self, ver_str):
        if not (ver := tuple(filter(str.isdigit, ver_str.split(".")))):
            raise Exception(f"Invalid version format!")
        ver = (*map(int, ver),)
        self.ver = ver + (0,) * (3 - len(ver))


VERSION = Version(*VERSION_TRIPLE)

BINDIR = os.path.dirname(ARGV[0])
LIBDIR = os.path.join(BINDIR, "lib")
CORE_DIR = os.path.join(BINDIR, "lib", "core")

if os.name == "nt":
    UNIX = False
else:
    UNIX = True

PYTHON_VER = sys.version
PYTHON_RAW_VER = (temp := sys.version_info).major, temp.minor, temp.micro

SYS_ARCH, EXE_FORM = platform.architecture()
EXE_FORM = EXE_FORM or constants.none
SYS_PROC = platform.processor() or constants.none
SYS_MACH = platform.machine()
SYS_INFO = platform.platform()
SYS_MACH_INFO = platform.uname()


def print_info():
    for name, value in globals().copy().items():
        if name in {"os", "sys", "print_info", "platform"} or not name.isupper():
            continue
        if not name.startswith("__") and not name.startswith("__"):
            print(f"{name} = {value!r}")


if __name__ == "__main__":
    print_info()



# Variable and Scope handling

import threading
import os
import sys
from . import constants
from . import info
from . import state
from . import error

# lock
W_LOCK = threading.Lock()
WS_LOCK = threading.Lock()

dependencies = {"dpl": set(), "python": {}}

# debug options
debug = {
    "show_instructions": 0,
    "show_scope_updates": 0,
    "show_value_updates": 0,
    "show_imports": 0,
    "warn_no_return": 0,
    "log_events": 0,
    "debug_output_file": "debug_log.txt",
    "track_time": 0,
    "time_threshold": 1.5,
    "_set_only_when_defined": 1,  # make sure that only defined variables in this scope can be set
}

flags = set()

# related to interpreter methods or behavior
# and meta programming to the extreme
meta = {
    "debug": debug,
    "argv": info.ARGV,
    "argc": info.ARGC,
    "inter_flags": flags,
    "internal": {
        "lib_path": info.LIBDIR,
        "main_path": "__main__",
        "main_file": constants.none,
        "version": info.VERSION,
        "raw_version": info.VERSION_TRIPLE,
        "pid": os.getpid(),
        "python_version": str(sys.version_info),
        "python_version_string": info.PYTHON_VER,
        "_set_only_when_defined": 1,
    },
    "dependencies": dependencies,
    "_set_only_when_defined": 1,
}

meta["internal"]["libs"] = {
    "core_libs": tuple(
        map(
            lambda x: os.path.basename(x),
            filter(
                os.path.isfile,
                (os.path.join(info.CORE_DIR, f) for f in os.listdir(info.CORE_DIR)),
            ),
        )
    ),
    "std_libs": tuple(
        map(
            lambda x: os.path.basename(x),
            filter(
                os.path.isfile,
                (os.path.join(info.LIBDIR, f) for f in os.listdir(info.LIBDIR)),
            ),
        )
    ),
}


def new_frame():
    "Generate a new scope frame"
    values_stack = []
    nscope(values_stack)
    return values_stack


def get_debug(name):
    "Get a debug option"
    return debug.get(name, None)


def is_debug_enabled(name):
    "Return a bool if a debug option is enabled"
    return bool(debug.get(name))


def set_debug(name, value):
    "Set a debug option"
    debug[name] = value


def nscope(frame):
    "New scope"
    t = {"_meta": meta, "_temp": {}}
    if frame:
        t["_global"] = frame[0]
        t["_nonlocal"] = frame[-1]
        t["_local"] = t
    else:
        t["_global"] = t
        t["_nonlocal"] = t
        t["_local"] = t
    with WS_LOCK:
        frame.append(t)
    if is_debug_enabled("show_scope_updates"):
        error.info(f"New scope created!")


def pscope(frame):
    "Pop the current scope also discarding"
    if len(frame) > 1:
        with WS_LOCK:
            frame.pop()
        if is_debug_enabled("show_scope_updates"):
            error.info(f"Scope discarded!")
    else:
        if is_debug_enabled("show_scope_updates"):
            error.info(f"Tried to discard global scope!")


def rget(dct, full_name, default=constants.nil, sep=".", meta=True):
    "Get a variable"
    if "." not in full_name:
        temp = dct.get(full_name, default)
        if is_debug_enabled("show_value_updates"):
            error.info(f"Variable {full_name!r} was read!")
        if meta and isinstance(temp, dict) and "[meta_value]" in temp:
            return temp["[meta_value]"]
        else:
            return temp
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if pos != last and name in node and isinstance(node[name], dict):
            node = node[name]
        elif pos == last and name in node:
            if is_debug_enabled("show_value_updates"):
                error.info(f"Variable {full_name!r} was read!")
            if meta and isinstance(node[name], dict) and "[meta_value]" in node[name]:
                return node[name]["[meta_value]"]
            else:
                return node[name]
        else:
            return default
    return default


def rpop(dct, full_name, default=constants.nil, sep="."):
    "Pop a variable"
    if "." not in full_name:
        with W_LOCK:
            temp = dct.get(full_name, default)
        return temp
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if pos != last and name in node and isinstance(node[name], dict):
            node = node[name]
        elif pos == last and name in node:
            if is_debug_enabled("show_value_updates"):
                error.info(f"Variable {full_name!r} was popped!")
            with W_LOCK:
                return node.pop(name)
        else:
            return default
    return default


def rset(dct, full_name, value, sep=".", meta=True):
    "Set a variable"
    if not isinstance(full_name, str):
        return full_name
    if "." not in full_name:
        with W_LOCK:
            if dct.get("_set_only_when_defined") and full_name not in dct:
                error.warn(
                    f"Tried to set {full_name!r} but scope was set to set only when defined."
                )
                return
            if (
                meta
                and isinstance(temp := dct.get("_const"), list)
                and full_name in temp
            ):
                return 1
            if meta and full_name in dct and "[update_mapping]" in (temp := dct):
                temp[temp["[update_mapping]"].get(full_name, full_name)] = value
                temp[full_name] = value
            if (
                meta
                and full_name in dct
                and isinstance((temp := dct[full_name]), dict)
                and "[meta_value]" in temp
            ):
                dct[full_name]["[meta_value]"] = value
                if "[update_mapping]" in temp:
                    vname = temp["[update_mapping]"].get("[onset]")
                    if vname:
                        temp[vname] = value
            else:
                dct[full_name] = value
            return
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if pos != last and name in node and isinstance(node[name], dict):
            node = node[name]
        elif pos == last:
            if node.get("_set_only_when_defined") and name not in node:
                error.warn(
                    f"Tried to set {full_name!r} but scope was set to set only when defined."
                )
                return
            with W_LOCK:
                if (
                    meta
                    and isinstance(temp := dct.get("_const"), list)
                    and name in temp
                ):
                    return 1
                if meta and name in node and "[update_mapping]" in (temp := node):
                    temp[temp["[update_mapping]"].get(name, name)] = value
                    temp[name] = value
                if (
                    meta
                    and name in node
                    and isinstance((temp := node[name]), dict)
                    and "[meta_value]" in temp
                ):
                    node[name]["[meta_value]"] = value
                    if "[update_mapping]" in temp:
                        vname = temp["[update_mapping]"].get("[onset]")
                        if vname:
                            temp[vname] = value
                else:
                    node[name] = value
            if is_debug_enabled("show_value_updates"):
                error.info(f"Variable {full_name!r} was set to `{value!r}`!")



from setuptools import setup, Extension
from Cython.Build import cythonize
import os

# Define the custom output path
output_dir = os.getcwd()

ext_modules = [Extension(name="parser", sources=["py_parser.py"])]

setup(
    ext_modules=cythonize(ext_modules, compiler_directives={"language_level": "3"}),
    zip_safe=False,
    # Add options for the build_ext command
    options={
        "build_ext": {
            "build_lib": output_dir,  # Set the custom output directory
            "force": True,  # Force recompilation of all files
        }
    },
)



# Parser and Preprocessor
# The heart, the interpreter of DPL

import os
import time
import sys
import itertools
import traceback
import threading
import pickle
from copy import deepcopy as copy

try:
    from . import arguments as argproc
    from . import info
    from . import state
    from . import error
    from . import utils
    from . import varproc
    from . import objects
    from . import constants
    from . import extension_support as ext_s
except ImportError:
    print(f"Please do not run it from here.")
    sys.exit(1)

IS_STILL_RUNNING = threading.Event()

dpl_func_attr = [
    "name",
    "body",
    "args",
    "docs",
    "defs",
]

threads = []
thread_events = (
    []
)  # Thread events, so any threads can be killed manually or automatically


def clean_threads():  # kill all threads and wait for them to terminate
    for i in thread_events:
        i.set()
    for i in threads:
        i.join()


def my_exit(code=0):
    IS_STILL_RUNNING.set()
    clean_threads()
    raise SystemExit(code)


sys.exit = my_exit
exit = my_exit

# setup runtime stuff. And yes on import.
try:
    import psutil

    CUR_PROCESS = psutil.Process()

    def get_memory(_, __):
        memory_usage = CUR_PROCESS.memory_info().rss
        return (utils.convert_bytes(memory_usage),)

    varproc.meta["internal"]["HasGetMemory"] = 1
    varproc.meta["internal"]["GetMemory"] = get_memory
except ModuleNotFoundError as e:
    varproc.meta["internal"]["HasGetMemory"] = 0
    varproc.meta["internal"]["GetMemory"] = lambda _, __: (state.bstate("nil"),)

varproc.meta["internal"].update(
    {
        "SetEnv": lambda _, __, name, value: os.putenv(name, value),
        "GetEnv": lambda _, __, name, default=None: os.getenv(name, default),
    }
)

varproc.meta["internal"]["os"] = {
    "uname": info.SYS_MACH_INFO,  # uname
    "architecture": info.SYS_ARCH,  # system architecture (commonly x86 or ARMv7 or whatever arm proc)
    "executable_format": info.EXE_FORM,  # name is self explanatory
    "machine": info.SYS_MACH,  # machine information
    "information": info.SYS_INFO,  # basically the tripple
    "processor": info.SYS_PROC,  # processor (intel and such)
    "threads": os.cpu_count(),  # physical thread count,
}


def get_size_of(_, __, object):
    return (utils.convert_bytes(sys.getsizeof(object)),)


try:
    get_size_of(0, 0, 0)
    varproc.meta["internal"]["SizeOf"] = get_size_of
except:

    def temp(_, __, ___):
        return f"err:{error.PYTHON_ERROR}:Cannot get memory usage of an object!\nIf you are using pypy, pypy does not support this feature."

    varproc.meta["internal"]["SizeOf"] = temp

varproc.meta["threading"] = {
    "runtime_event": IS_STILL_RUNNING,
    "is_still_running": lambda _, __: IS_STILL_RUNNING.is_set(),
}

varproc.meta["str_intern"] = lambda _, __, string: sys.intern(string)


def get_block(code, current_p):
    "Get a code block"
    pos, file, _, _ = code[current_p]
    p = current_p + 1
    k = 1
    res = []
    while p < len(code):
        _, _, ins, _ = code[p]
        if ins in info.INC_EXT:
            k += 1
        elif ins in info.INC:
            k -= info.INC[ins]
        elif ins in info.DEC:
            k -= 1
        if k == 0:
            break
        else:
            res.append(code[p])
        p += 1
    else:
        print(f"Error in line {pos} file {file!r}\nCause: Block wasnt closed!")
        return None
    return p, res


def get_cust(code, current_p, INC, INC_EXT, DEC):
    "Get a code block"
    pos, file, _, _ = code[current_p]
    p = current_p + 1
    k = 1
    res = []
    while p < len(code):
        _, _, ins, _ = code[p]
        if ins in INC_EXT:
            k += 1
        elif ins in INC:
            k -= info.INC[ins]
        elif ins in DEC:
            k -= 1
        if k == 0:
            break
        else:
            res.append(code[p])
        p += 1
    else:
        print(f"Error in line {pos} file {file!r}\nCause: Block wasnt closed!")
        return None
    return p, res


def has(attrs, dct):
    return True if False not in map(lambda x: x in dct, attrs) else False


def process(code, name="__main__"):
    "Preprocess a file"
    res = []
    nframe = varproc.new_frame()
    dead_code = True
    warnings = True
    define_func = False
    for lpos, line in filter(
        lambda x: (
            True
            if x[1] and not x[1].startswith("#") and not x[1].startswith("...")
            else False
        ),
        enumerate(map(str.strip, code.split("\n")), 1),
    ):
        if line.startswith("&"):
            ins, *args = line[1:].lstrip().split()
            args = argproc.bs_thing(nframe, args)
            argc = len(args)
            if ins == "include" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.join(info.LIBDIR, args[0][1:-1])
                elif args[0].startswith('"') and args[0].endswith('"'):
                    file = os.path.join(os.path.dirname(name), args[0][1:-1])
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), file)
                if not os.path.isfile(file):
                    print("File not found:", file)
                    break
                with open(file, "r") as f:
                    res.extend(process(f.read(), name=file))
                file = os.path.realpath(file)
                varproc.meta["dependencies"]["dpl"].add(file)
            elif ins == "set_name" and argc == 1:
                name = str(args[0])
            elif ins == "includec" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.join(info.LIBDIR, args[0][1:-1])
                elif args[0].startswith('"') and args[0].endswith('"'):
                    file = os.path.join(os.path.dirname(name), args[0][1:-1])
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), file)
                else:
                    assert False, "This is not reachable!"
                if not os.path.isfile(file):
                    print("File not found:", file)
                    break
                with open(file, "rb") as f:
                    res.extend(pickle.loads(f.read()))
                file = os.path.realpath(file)
                varproc.meta["dependencies"]["dpl"].add(file)
            elif ins == "extend" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.join(info.LIBDIR, args[0][1:-1])
                elif args[0].startswith('"') and args[0].endswith('"'):
                    file = os.path.join(os.path.dirname(name), args[0][1:-1])
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), file)
                else:
                    assert False, "This is not reachable!"
                if not os.path.isfile(file):
                    print("File not found:", file)
                    break
                with open(file, "r") as f:
                    res.extend(process(f.read(), name=name))
                file = os.path.realpath(file)
                varproc.meta["dependencies"]["dpl"].add(file)
            elif ins == "use" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.join(info.LIBDIR, (ofile := args[0][1:-1]))
                    search_path = "_std"
                elif args[0].startswith('"') and args[0].endswith('"'):
                    file = os.path.join(os.path.dirname(name), (ofile := args[0][1:-1]))
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), file)
                    search_path = "_loc"
                else:
                    assert False, "This is not reachable!"
                if not os.path.isfile(file):
                    print("File not found:", file)
                    break
                if ext_s.py_import(nframe, file, search_path, loc="."):
                    print(f"Something wrong happened...")
                    return error.PREPROCESSING_ERROR
                ofile = os.path.realpath(file)
                if search_path in varproc.meta["dependencies"]["python"]:
                    varproc.meta["dependencies"]["python"][search_path].add(ofile)
                else:
                    varproc.meta["dependencies"]["python"][search_path] = {ofile}
            elif ins == "version" and argc == 1:
                if err := info.VERSION.getDiff(args[0]):
                    error.pre_error(lpos, name, f"{name!r}:{lpos}: {err}")
                    return error.COMPAT_ERROR
            elif ins == "dead_code_disable" and argc == 0:
                dead_code = False
            elif ins == "dead_code_enable" and argc == 0:
                dead_code = True
            elif ins == "warn_code_disable" and argc == 0:
                warnings = False
            elif ins == "warn_code_enable" and argc == 0:
                warnings = True
            elif ins == "def_fn_disable" and argc == 0:
                define_func = False
            elif ins == "def_fn_enable" and argc == 0:
                define_func = True
            else:
                error.pre_error(
                    lpos, name, f"{name!r}:{lpos}: Invalid directive {ins!r}"
                )
                break
        else:
            if " " in line:
                ins, arg = line.strip().split(maxsplit=1)
                args = argproc.exprs_preruntime(argproc.group(arg))
            else:
                ins = line
                args = []
            res.append((lpos, name, ins, args))
    else:
        if dead_code and info.DEAD_CODE_OPT:
            p = 0
            warn_num = 0
            nres = []
            while p < len(res):
                line = pos, file, ins, args = res[p]
                if (
                    ins in {"for", "loop", "while", "thread"}
                    and p + 1 < len(res)
                    and res[p + 1][2] in {"end", "stop", "skip"}
                ):
                    if warnings and info.WARNINGS:
                        error.warn(
                            f"Warning: {ins!r} statement is empty!\nLine {pos}\nIn file {file!r}"
                        )
                    temp = get_block(res, p)
                    if temp:
                        p, _ = temp
                    else:
                        return []
                    warn_num += 1
                elif (
                    ins in {"if", "if_else", "module"}
                    and p + 1 < len(res)
                    and res[p + 1][2] == "end"
                ):
                    if warnings and info.WARNINGS:
                        error.warn(
                            f"Warning: {ins!r} statement is empty!\nLine {pos}\nIn file {file!r}"
                        )
                    temp = get_block(res, p)
                    if temp:
                        p, _ = temp
                    else:
                        return []
                    if ins == "if_else":
                        temp = get_block(res, p)
                        if temp:
                            p, _ = temp
                        else:
                            return []
                    warn_num += 1
                elif (
                    ins in {"fn", "method", "body"}
                    and p + 1 < len(res)
                    and res[p + 1][2] in {"end", "return"}
                ):
                    if len(args) == 0:
                        error.warn(
                            f"Error: Malformed function definition!\nLine {pos}\nIn file {file!r}"
                        )
                    if warnings and info.WARNINGS:
                        error.warn(
                            f"Warning: Function {line[3][0]!r} is empty!\nLine {pos}\nIn file {file!r}"
                        )
                    temp = get_block(res, p)
                    if temp:
                        p, _ = temp
                    else:
                        return []
                    if define_func:
                        if warnings and info.WARNINGS:
                            print(
                                f'Warning: set "{line[3][0]}" none\nLine {pos}\nIn file {file!r}'
                            )
                        nres.append(
                            (pos, file, "set", [f'"{line[3][0]}"', constants.none])
                        )
                        warn_num += 1
                    warn_num += 1
                else:
                    nres.append(line)
                p += 1
            if warnings and info.WARNINGS and warn_num:
                print(f"Warning Info: {warn_num:,} Total warnings.")
        return {
            "code": nres if dead_code and info.DEAD_CODE_OPT else res,
            "frame": nframe or None,
        }
    return error.PREPROCESSING_ERROR


def run(code, frame=None, thread_event=IS_STILL_RUNNING, generator_pc=None):
    "Run code generated by 'process'"
    p = 0
    end_time = start_time = 0
    if isinstance(code, dict):
        code, nframe = code["code"], code["frame"]
    elif isinstance(code, int):
        return code
    else:
        nframe = {}
    sys.stdout.flush()
    if frame is not None:
        frame[0].update(nframe[0])
    else:
        frame = nframe
    varproc.rset(frame[-1], "_generator_process_communication", generator_pc)
    while p < len(code) and not IS_STILL_RUNNING.is_set():
        pos, file, ins, args = code[p]
        if ins not in {  # Lazy evaluation
            "while",
        }:
            try:
                args = argproc.exprs_runtime(frame, args)
            except Exception as e:
                error.error(
                    pos,
                    file,
                    f"Something went wrong when arguments were processed:\n{traceback.format_exc()}\n> {args!r}",
                )
                return error.PYTHON_ERROR
        if varproc.is_debug_enabled("show_instructions"):
            error.info(f"Executing: {code[p]}")
        argc = len(args)
        if ins == "fn" and argc >= 1:
            name, *params = args
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            varproc.rset(frame[-1], name, objects.make_function(name, body, params))
        elif ins == "for" and argc == 3 and args[1] == "in":
            name, _, iter = args
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            if body:
                for i in iter:
                    frame[-1][name] = i
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "for" and argc == 4 and args[2] == "in":
            pos_name, name, _, iter = args
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            if body:
                for posv, i in enumerate(iter):
                    frame[-1][name] = i
                    frame[-1][pos_name] = posv
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "loop" and argc == 0:
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            if body:
                while not thread_event.is_set():
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "loop" and argc == 1:
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            if body:
                for _ in range(args[0]):
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "while" and argc != 0:
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            if body:
                while not thread_event.is_set():
                    try:
                        (res,) = argproc.exprs_runtime(frame, args)
                        if not res:
                            break
                    except Exception as e:
                        error.error(
                            pos,
                            file,
                            f"Something went wrong when arguments were processed:\n{e}\n> {args!r}",
                        )
                        return error.RUNTIME_ERROR
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "stop" and argc == 0:
            return error.STOP_RESULT
        elif ins == "skip" and argc == 0:
            return error.SKIP_RESULT
        elif ins == "if" and argc == 1:
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            if args[0]:
                err = run(body, frame=frame)
                if err:
                    return err
        elif ins == "if_else" and argc == 1:
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, true = temp
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, false = temp
            if args[0]:
                err = run(true, frame=frame)
                if err:
                    return err
            else:
                err = run(false, frame=frame)
                if err:
                    return err
        elif ins == "set" and argc == 2:
            if t := varproc.rset(frame[-1], args[0], args[1]):
                error.error(
                    pos,
                    file,
                    f"Tried to set a constant variable!\nPlease use fset instead!\nLine {pos}\nFile {file}",
                )
                return error.NAME_ERROR
        elif ins == "const" and argc == 2:
            name = args[0]
            if varproc.rset(frame[-1], name, args[1]):
                error.error(
                    pos,
                    file,
                    "Tried to set a constant variable!\nPlease use fset instead!\nLine {pos}\nFile {file}",
                )
                return error.RUNTIME_ERROR
            consts = frame[-1].get("_const")
            if consts:
                consts.append(name)
            else:
                frame[-1]["_const"] = [name]
        elif ins == "fset" and argc == 2:
            varproc.rset(frame[-1], args[0], args[1], meta=False)
        elif ins == "del" and argc == 1:
            varproc.rpop(frame[-1], args[0])
            consts = frame[-1].get("_const")
            if consts and name in consts:
                consts.remove(name)
        elif ins == "expect" and argc == 1:
            vname = args[0]
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            err = run(body, frame)
            err_name = error.ERRORS_DICT.get(err, err)
            if isinstance(err_name, int):
                ...
            else:
                err = (err_name, err)
            varproc.rset(frame[-1], vname, err)
        elif ins == "raise" and argc == 1 and isinstance(args[0], int):
            return args[0]
        elif ins == "module" and argc == 1:
            name = args[0]
            temp = [frame[-1]]
            varproc.nscope(temp)
            btemp = get_block(code, p)
            if btemp == None:
                break
            else:
                p, body = btemp
            err = run(body, temp)
            if err:
                return err
            varproc.rset(frame[-1], name, temp[1])
            del temp
        elif ins == "object" and argc == 1:
            varproc.rset(frame[-1], args[0], objects.make_object(args[0]))
        elif ins == "new" and argc == 2:
            obj = args[0]
            if obj == state.bstate("nil"):
                error.error(pos, file, f"Unknown object")
                break
            varproc.rset(obj, "_internal.name", args[1])
            varproc.rset(frame[-1], args[1], copy(obj))
        elif ins == "method" and argc >= 2:
            self, name, *params = args
            if self == state.bstate("nil"):
                error.error(
                    pos, file, "Cannot bind a method to a value that isnt a context!"
                )
                return error.RUNTIME_ERROR
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            varproc.rset(self, name, objects.make_method(name, body, params, self))
        elif ins == "START_TIME" and argc == 0:
            start_time = time.perf_counter()
        elif ins == "STOP_TIME" and argc == 0:
            end_time = time.perf_counter() - start_time
        elif ins == "LOG_TIME" and argc == 0:
            ct, unit = utils.convert_sec(end_time)
            error.info(f"Elapsed time: {ct:,.8f}{unit}")
        elif ins == "LOG_TIME" and argc == 1:
            ct, unit = utils.convert_sec(end_time)
            error.info(f"Elapsed time: {args[0]} {ct:,.8f}{unit}")
        elif ins == "cmd" and argc == 1:
            os.system(args[0])
        elif ins == "pass":
            ...
        elif ins == "thread" and argc == 0:
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp

            def th():
                if err := run(body, frame, thread_event):
                    raise RuntimeError(f"Thread returned an error: {err}")

            th_obj = threading.Thread(target=th)
            threads.append(th_obj)
            th_obj.start()
        elif ins == "new_thread_event" and argc == 1:
            varproc.rset(frame[-1], args[0], (temp := threading.Event()))
            thread_events.append(temp)
        elif ins == "thread" and argc == 1:
            if not isinstance(args[0], threading.Event):
                error.error(pos, file, "The given thread event was invalid!")
                return error.THREAD_ERROR
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp

            def th():
                if err := run(body, frame, args[0]):
                    raise RuntimeError(f"Thread returned an error: {err}")

            th_obj = threading.Thread(target=th)
            threads.append(th_obj)
            th_obj.start()
        elif ins == "exit" and argc == 0:
            my_exit()
        elif ins == "return":  # Return to the latched names
            if not (temp := varproc.rget(frame[-1], "_returns")) != state.bstate("nil"):
                ...
            else:
                for name, value in zip(temp, args):
                    varproc.rset(frame[-1], f"_nonlocal.{name}", value)
                if (tmp := frame[-1].get("_memoize")) not in constants.constants_false:
                    tmp[0][tmp[1]] = tuple(
                        map(
                            lambda x: (
                                x
                                if isinstance(x, (str, int, float, tuple, complex))
                                else f"{type(x)}:{id(x)}"
                            ),
                            args,
                        )
                    )
            return error.STOP_RESULT
        elif (
            ins == "freturn"
        ):  # Return to the latched names with no memoization detection (faster)
            if not (temp := varproc.rget(frame[-1], "_returns")) != state.bstate("nil"):
                ...
            else:
                for name, value in zip(temp, args):
                    varproc.rset(frame[-1], f"_nonlocal.{name}", value)
            return error.STOP_RESULT
        elif ins == "help" and argc == 1:
            if not isinstance(args[0], dict) and hasattr(args[0], "__doc__"):
                doc = getattr(args[0], "__doc__")
                if doc:
                    print(
                        f"\nHelp on {getattr(args[0], '__name__', '???')}, line [{pos}]:\n{doc}"
                    )
                else:
                    help(args[0])
            elif not isinstance(args[0], dict):
                return error.TYPE_ERROR
            else:
                temp = varproc.rget(
                    args[0], "docs", default=varproc.rget(args[0], "_internal.docs")
                )
                if temp == state.bstate("nil"):
                    print(f"\nHelp, line [{pos}]: No documentation was found!")
                else:
                    print(f"\nHelp, line [{pos}]:\n{temp}")
        elif ins == "wait_for_threads" and argc == 0:
            for i in threads:
                i.join()
            threads.clear()
        elif ins == "catch" and argc >= 2:  # catch return value of a function
            rets, func_name, *args = args
            if (temp := varproc.rget(frame[-1], func_name)) == state.bstate(
                "nil"
            ) or not isinstance(temp, dict):
                error.error(pos, file, f"Invalid function {func_name!r}!")
                break
            varproc.nscope(frame)
            if temp["defs"]:
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        frame[-1][name] = temp["defs"].get(name, state.bstate("nil"))
                    else:
                        frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    varproc.rset(frame[-1], name, value)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            frame[-1]["_returns"] = rets
            err = run(temp["body"], frame)
            if err > 0:
                return err
            varproc.pscope(frame)
        elif ins == "mcatch" and argc >= 2:  # catch return value of a function
            rets, func_name, *args = args
            mem_args = tuple(
                map(
                    lambda x: (
                        x
                        if isinstance(x, (str, int, float, tuple, complex))
                        else f"{type(x)}:{id(x)}"
                    ),
                    args,
                )
            )
            if (
                (temp := varproc.rget(frame[-1], func_name)) == state.bstate("nil")
                and isinstance(temp, dict)
                and mem_args in temp
            ):
                error.error(pos, file, f"Invalid function {func_name!r}!")
                break
            if mem_args in temp["memoize"]:
                for name, value in zip(rets, temp["memoize"][mem_args]):
                    varproc.rset(frame[-1], name, value)
                p += 1
                continue
            varproc.nscope(frame)
            if temp["defs"]:
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        frame[-1][name] = temp["defs"].get(name, state.bstate("nil"))
                    else:
                        frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    varproc.rset(frame[-1], name, value)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            frame[-1]["_returns"] = rets
            frame[-1]["_memoize"] = (temp["memoize"], mem_args)
            err = run(temp["body"], frame)
            if err > 0:
                return err
            varproc.pscope(frame)
        elif ins == "body" and argc >= 1:  # give a code block to a python function
            name, *args = args
            if (temp := varproc.rget(frame[-1], name)) == state.bstate(
                "nil"
            ) or not hasattr(temp, "__call__"):
                error.error(pos, file, f"Invalid function {name!r}!")
                break
            try:
                btemp = get_block(code, p)
                if btemp is None:
                    break
                else:
                    p, body = btemp
                if argc == 2 and isinstance(args[0], dict) and args[0].get("[KWARGS]"):
                    args[0].pop("[KWARGS]")
                    pa = args[0].pop("[PARGS]", tuple())
                    res = ext_s.call_w_body(
                        temp,
                        frame,
                        varproc.meta["internal"]["main_path"],
                        body,
                        pa,
                        args[0],
                    )
                else:
                    res = ext_s.call_w_body(
                        temp, frame, varproc.meta["internal"]["main_path"], body, args
                    )
                if isinstance(res, tuple):
                    for name, value in zip(rets, res):
                        varproc.rset(frame[-1], name, value)
                elif isinstance(res, int) and res:
                    return res
                elif isinstance(res, str):
                    if res == "err":
                        break
                    elif res == "stop":
                        return error.STOP_RESULT
                    elif res == "skip":
                        return error.SKIP_RESULT
                    elif res.startswith("err:"):
                        _, ecode, message = res.split(":", 2)
                        error.error(pos, file, message)
                        return int(ecode)
            except:
                error.error(pos, file, traceback.format_exc()[:-1])
                return error.PYTHON_ERROR
        elif ins == "pause" and argc == 0:
            input()
        elif ins == "raise" and argc in (0, 1, 2):
            if argc == 0:
                return error.RUNTIME_ERROR
            elif argc == 1:
                return args[0]
            elif argc == 2:
                error.error(file, pos, args[1])
                return args[0]
            else:
                error.error(file, pos, "Invalid raise statement!")
                break
        elif ins == "pycatch" and argc >= 2:  # catch return value of a python function
            rets, name, *args = args
            if (temp := varproc.rget(frame[-1], name)) == state.bstate(
                "nil"
            ) or not hasattr(temp, "__call__"):
                error.error(pos, file, f"Invalid function {name!r}!")
                return error.NAME_ERROR
            try:
                if argc == 3 and isinstance(args[0], dict) and args[0].get("[KWARGS]"):
                    args[0].pop("[KWARGS]")
                    pa = args[0].pop("[PARGS]", tuple())
                    res = ext_s.call(
                        temp, frame, varproc.meta["internal"]["main_path"], pa, args[0]
                    )
                else:
                    res = ext_s.call(
                        temp, frame, varproc.meta["internal"]["main_path"], args
                    )
                if (
                    res is None
                    and info.WARNINGS
                    and varproc.is_debug_enabled("warn_no_return")
                ):
                    error.warn(
                        "Function doesnt return anything. To reduce overhead please dont use pycatch.\nLine {pos}\nFile {file}"
                    )
                if isinstance(res, tuple):
                    for name, value in zip(rets, res):
                        varproc.rset(frame[-1], name, value)
                elif isinstance(res, int) and res:
                    return res
                elif isinstance(res, str):
                    if res == "err":
                        break
                    elif res == "stop":
                        return error.STOP_RESULT
                    elif res == "skip":
                        return error.SKIP_RESULT
                    elif res.startswith("err:"):
                        _, ecode, message = res.split(":", 2)
                        error.error(pos, file, message)
                        return int(ecode)
            except:
                error.error(pos, file, traceback.format_exc()[:-1])
                return error.PYTHON_ERROR
        elif ins == "template" and argc == 1:
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            dct = {}
            for _, _, vname, vtype in body:
                if not vtype:
                    dct[vname] = state.bstate("types:any")
                else:
                    (dct[vname],) = argproc.bs_thing(frame, vtype)
            varproc.rset(frame[-1], args[0], dct)
        elif ins == "from_template" and argc == 2:
            template = args[0]
            tname = args[0]
            temp = get_block(code, p)
            if temp is None:
                break
            else:
                p, body = temp
            dct = {}
            if template != constants.none:
                for pos, _, vname, vitem in body:
                    if vname in dct:
                        error.error(
                            pos, file, f"Attribute {vname!r} is already defined!"
                        )
                        return error.NAME_ERROR
                    if vitem == ["$name"]:
                        value = args[1]
                    else:
                        (value,) = argproc.bs_thing(frame, vitem)
                    if vname not in template:
                        error.error(
                            pos,
                            file,
                            f"Attribute {vname!r} is not in the defined template.",
                        )
                        return error.NAME_ERROR
                    if template[vname] != state.bstate("types:any") and not isinstance(
                        value, template[vname]
                    ):
                        error.error(
                            pos,
                            file,
                            f"Attribute {vname!r} must be type {template[vname]!r}! Not {value!r} of {type(value)!r}",
                        )
                        return error.TYPE_ERROR
                    dct[vname] = value
                for i in template.keys():
                    if i not in dct:
                        dct[i] = constants.none
            else:
                for pos, _, vname, vitem in body:
                    (dct[vname],) = argproc.bs_thing(frame, vitem)
            varproc.rset(frame[-1], args[1], dct)
        elif (
            (temp := varproc.rget(frame[-1], ins, default=varproc.rget(frame[0], ins)))
            != state.bstate("nil")
            and isinstance(temp, dict)
            and has(dpl_func_attr, temp)
        ):  # Call a function
            varproc.nscope(frame)
            if temp["defs"]:
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        frame[-1][name] = temp["defs"].get(name, state.bstate("nil"))
                    else:
                        frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    varproc.rset(frame[-1], name, value)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            err = run(temp["body"], frame)
            if err:
                return err
            varproc.pscope(frame)
        elif (
            temp := varproc.rget(frame[-1], ins, default=varproc.rget(frame[0], ins))
        ) != state.bstate("nil") and hasattr(
            temp, "__call__"
        ):  # call a python function
            try:
                if argc == 1 and isinstance(args[0], dict) and args[0].get("[KWARGS]"):
                    args[0].pop("[KWARGS]")
                    pa = args[0].pop("[PARGS]", tuple())
                    res = ext_s.call(
                        temp, frame, varproc.meta["internal"]["main_path"], pa, args[0]
                    )
                else:
                    res = ext_s.call(
                        temp, frame, varproc.meta["internal"]["main_path"], args
                    )
                if isinstance(res, int) and res:
                    return res
                elif isinstance(res, str):
                    if res == "break":
                        break
                    elif res.startswith("err:"):
                        _, ecode, message = res.split(":", 2)
                        error.error(pos, file, message)
                        return int(ecode)
            except:
                error.error(pos, file, traceback.format_exc()[:-1])
                return error.PYTHON_ERROR
        else:
            if not isinstance((obj := varproc.rget(frame[-1], ins)), dict) and obj in (
                None,
                constants.none,
            ):
                print(
                    "\nAdditional Info: User may have called a partially defined function!",
                    end="",
                )
            error.error(pos, file, f"Invalid instruction {ins}\n{args}")
            return error.RUNTIME_ERROR
        p += 1
    else:
        return 0
    error.error(pos, file, "Error was raised!")
    return error.SYNTAX_ERROR


# to avoid circular imports
ext_s.register_run(run)
ext_s.register_process(process)



import os


def run(python_bin):
    print(f"Building with {python_bin}")
    local = os.path.dirname(__file__)
    if os.path.isfile(temp := os.path.join(local, "py_parser.c")):
        os.remove(temp)

    print(f"Building parser.{os}")

    os.chdir(local)
    os.system(f"{python_bin} setup.py build_ext --inplace")



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

types = dpl.extension("types")
types["int"] = int
types["str"] = str
types["flt"] = float
types["list"] = list
types["dict"] = dict
types["tuple"] = tuple
types["complex"] = complex
types["any"] = dpl.state.bstate("types:any")
types["Exception"] = Exception



# Simple package installer!

import requests
import os, sys
import zipfile
import shutil
import socket


def dl_repo(user_name, repo_name, branch="master", use_branch_name=True, location="."):
    zip_url = (
        f"https://github.com/{user_name}/{repo_name}/archive/refs/heads/{branch}.zip"
    )

    try:
        print("Validating connection...")
        socket.create_connection(("8.8.8.8", 53), timeout=5)
    except Exception as e:
        print("Must be connected to the internet!")
        return

    destination_dir = os.path.join(
        location, f"{repo_name}-{branch}" if use_branch_name else repo_name
    )
    if os.path.exists(destination_dir):
        if os.path.isfile(destination_dir):
            print(
                "Error: A file has the same name as the repo. Please rename the file."
            )
            return
        else:
            print(
                "A directory has the same name.\nThis will be treated as a reinstallation."
            )
            if input("Enter [y] to continue: ").lower() not in {"y", "yes"}:
                return
            print(f"{destination_dir}: Deleting the directory...")
            shutil.rmtree(destination_dir)
    print(f"{repo_name}: Downloading zip...")
    try:
        response = requests.get(zip_url)
    except Exception as e:
        print(e)
        return
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)
    print(f"{repo_name}: Done!")
    if response.status_code == 200:
        zip_path = os.path.join(destination_dir, f"{repo_name}-{branch}.zip")
        print(f"{repo_name}: Writing to {destination_dir}.zip...")
        with open(zip_path, "wb") as file:
            file.write(response.content)

        print(f"{destination_dir}: Extracting...")
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(destination_dir)

        inner_folder = os.path.join(destination_dir, f"{repo_name}-{branch}")

        print(f"{inner_folder}: Unwrapping...")
        if os.path.isdir(inner_folder):
            for item in os.listdir(inner_folder):
                item_path = os.path.join(inner_folder, item)
                shutil.move(item_path, os.path.join(destination_dir, item))

            print(f"{inner_folder}: Deleting inner folder...")
            os.rmdir(inner_folder)

        print(f"{destination_dir}: Deleting zipfile...")
        os.remove(zip_path)
        print(f"Done!\nInstalled as: {destination_dir}")
    else:
        print(f"Failed to download repository. Status code: {response.status_code}")


def delete(path):
    shutil.rmtree(path)


def main():
    if len(sys.argv) != 2:
        print("Invalid invocation!\nFor help: dfpm help")
        return 1
    elif sys.argv[1] == "help":
        print(
            "DFPM (DPLs Friendly Package Manager) v0.0.1\nFormat: https://github.com/<user>/<repo>=<branch>"
        )
    elif "=" not in sys.argv[0]:
        print("Format: https://github.com/<user>/<repo>=<branch>\nFor more: dfpm help")
        return 1
    else:
        repo_url, branch_name = sys.argv[1].split("=", maxsplit=1)
        dl_repo(repo_url, branch_name)
    return 0


if __name__ == "__main__":
    sys.exit(main())



from copy import deepcopy as copy


def flags(argv: list[str], remove_first=False) -> tuple:
    if remove_first:
        argv.pop(0)
    indexes: dict[int, str] = {}
    end = 0
    for pos, value in enumerate(argv):
        if value.startswith("-"):
            if pos > end:
                end = pos
    for pos, value in enumerate(argv):
        if value.startswith("--"):
            if "=" not in value:
                continue
            indexes[pos] = value[2:]
        elif value.startswith("-"):
            indexes[pos] = value[1:]
        else:
            break
    for i in indexes.keys():
        if i <= end:
            argv.pop(min(indexes.keys()))
    return tuple(set(map(lambda x: x[1], indexes.items())))



# A simple way of making objects and methods n stuff...

from . import varproc
from . import constants

id_tracker = 0


def set_repr(frame, name="???", type_name=None, repr=None):
    global id_tracker
    if "_internal" not in frame:
        frame["_internal"] = {
            "name": name,
            "type": f"type:{type_name or name}",
            "docs": "An object.",
            "id": id_tracker,
        }
    if "_im_repr" not in frame:
        frame["_im_repr"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defs": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (repr or f"<{type_name or 'Object'} {name} ID: {id_tracker}>",),
                )
            ],
        }
    id_tracker += 1
    return frame


def make_function(name, body, params):
    global id_tracker
    id_tracker += 1
    return set_repr(
        {
            "name": name,
            "body": body,
            "args": params,
            "self": constants.nil,
            "docs": f"Function. ({name!r})",
            "defs": {},
            "memoize": {},
            "id": id_tracker - 1,
        },
        name,
        "builtin-function-object",
    )


def make_method(name, body, params, self):
    global id_tracker
    id_tracker += 1
    return set_repr(
        {
            "name": name,
            "body": body,
            "args": params,
            "self": self,
            "docs": f"Method of {varproc.rget(self, '_internal.name')}. ({name})",
            "defs": {},
            "id": id_tracker - 1,
        },
        name,
        "builtin-method-object",
    )


def make_object(name):
    global id_tracker
    id_tracker += 1
    return set_repr(
        {
            "_internal": {
                "name": name,
                "type": f"type:{name}",
                "docs": f"An object. ({name})",
                "id": id_tracker - 1,
            }
        },
        name,
        "builtin-object:Object",
    )



# Used to handle arguments and expressions
# NOT FOR THE CLI

from . import state
from . import constants
from . import error
from . import varproc
from .info import *

# Functions in utils that ciuldnt be imported


def flatten_dict(d, parent_key="", sep=".", seen=None):
    if seen is None:
        seen = set()
    items = {}
    dict_id = id(d)
    if dict_id in seen:
        return d
    seen.add(dict_id)
    for key, value in d.items():
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(flatten_dict(value, new_key, sep, seen))
        elif isinstance(value, list):
            items[f"{new_key}"] = value
            for i, item in enumerate(value):
                items[f"{new_key}[{i}]"] = item
        else:
            items[new_key] = value
    seen.remove(dict_id)
    return items


# Use to make the argument handler to handle
# char literals
class char:
    def __init__(self, value):
        self.val = value

    def __repr__(self):
        return self.val


methods = {}


# Need I explain?
def is_int(arg):
    return arg.replace("-", "").replace(",", "").isdigit()


def is_hex(arg):
    if not arg.startswith("0x"):
        return False
    try:
        int(arg[2:], 16)
        return True
    except:
        return False


def is_bin(arg):
    if not arg.startswith("0b"):
        return False
    try:
        int(arg[2:], 2)
        return True
    except:
        return False


def is_float(arg):
    return arg.replace("-", "").replace(",", "").replace(".", "").isdigit()


def is_id(arg):
    return arg.replace(".", "").replace("_", "").isalnum()


def is_sid(arg):
    return arg.startswith("name{") and arg.endswith("}")


def is_var(arg):
    return arg.startswith("%") and is_id(arg[1:])


def is_fvar(arg):
    return arg.startswith(":") and is_id(arg[1:])


def is_svar(arg):
    return arg.startswith("%{") and arg.endswith("}")


def is_sfvar(arg):
    return arg.startswith(":{") and arg.endswith("}")


def expr_preruntime(arg):
    "Process arguments at preprocessing"
    if not isinstance(arg, str):
        return arg
    elif is_int(arg):
        return int(arg.replace(",", ""))
    elif is_float(arg.replace(",", "")):
        return float(arg)
    elif is_bin(arg):
        return int(arg, 2)
    elif is_hex(arg):
        return int(arg, 16)
    elif arg == "true":
        return constants.true
    elif arg == "false":
        return constants.false
    elif arg == "none":
        return constants.none
    elif arg == "nil":
        return constants.nil
    elif arg == "...":
        return constants.elipsis
    elif arg == "[]":
        return []
    return arg


def expr_runtime(frame, arg):
    "Process an argument at runtime"
    if isinstance(arg, char):
        return arg.val
    elif not isinstance(arg, str):
        return arg
    if is_id(arg):
        return arg
    elif is_var(arg):
        return varproc.rget(frame[-1], arg[1:], default=varproc.rget(frame[0], arg[1:]))
    elif is_fvar(arg):
        return varproc.rget(
            frame[-1],
            arg[1:],
            default=varproc.rget(frame[0], arg[1:], meta=False),
            meta=False,
        )
    elif is_svar(arg):
        return varproc.rget(
            frame[-1], arg[2:-1], default=varproc.rget(frame[0], arg[2:-1])
        )
    elif is_sfvar(arg):
        return varproc.rget(
            frame[-1],
            arg[2:-1],
            default=varproc.rget(frame[0], arg[2:-1], meta=False),
            meta=False,
        )
    elif is_sid(arg):
        return arg[5:-1]
    else:
        return expr_preruntime(arg)


def add_method(name=None, from_func=False, process=True):
    def wrapper(func):
        fname = name if name is not None else getattr(func, "__name__", "_dump")
        methods[fname] = (
            lambda *arg: func(None, None, *arg) if from_func else func,
            process,
        )
        return func

    return wrapper


def my_range(start, end):
    def pos(start, end):
        while start < end:
            yield start
            start += 1

    def neg(start, end):
        while start > end:
            yield start
            start -= 1

    return pos(start, end) if start < end else neg(start, end)


def evaluate(frame, expression):
    "Evaluate an expression"
    ins, *args = expression
    if ins in methods:
        if methods[ins][0]:
            args = bs_thing(frame, args)
        return methods[ins][1](frame, *args)
    match (expression):
        case ["Range", arg]:
            arg = express(frame, arg)
            return tuple(range(arg))
        case ["dRange", start, end]:
            start = express(frame, start)
            end = express(frame, end)
            return tuple(my_range(start, end))
        case ["dRawRange", start, end]:
            start = express(frame, start)
            end = express(frame, end)
            return my_range(start, end)
        case ["RawRange", arg]:
            arg = express(frame, arg)
            return range(arg)
        case ["Sum", *items]:
            items = exprs_runtime(frame, items)
            start_t = type(items[0])
            start = start_t()
            for i in items:
                try:
                    start += i
                except:
                    start += start_t(i)
            return start
        case ["Index", lst, index]:
            return express(frame, lst)[express(frame, index)]
        case [op1, "+", op2]:
            return express(frame, op1) + express(frame, op2)
        case [op1, "..", op2]:
            return str(express(frame, op1)) + str(express(frame, op2))
        case [op1, "-", op2]:
            return express(frame, op1) - express(frame, op2)
        case [op1, "*", op2]:
            return express(frame, op1) * express(frame, op2)
        case [op1, "/", op2]:
            return express(frame, op1) / express(frame, op2)
        case [op1, "%", op2]:
            return express(frame, op1) % express(frame, op2)
        case [op1, "^", op2]:
            return express(frame, op1) ** express(frame, op2)
        case [op1, "caseless{==}", *op2]:
            express(frame, op2)[0].lower()
            return (
                constants.true
                if express(frame, op1).lower() == bs_thing(frame, op2)[0].lower()
                else constants.false
            )
        case [op1, "caseless{!=}", *op2]:
            return (
                constants.true
                if express(frame, op1).lower() != bs_thing(frame, op2)[0].lower()
                else constants.false
            )
        case [op1, "caseless{>}", *op2]:
            return (
                constants.true
                if express(frame, op1).lower() > bs_thing(frame, op2)[0].lower()
                else constants.false
            )
        case [op1, "caseless{<}", *op2]:
            return (
                constants.true
                if express(frame, op1).lower() < bs_thing(frame, op2)[0].lower()
                else constants.false
            )
        case [op1, "caseless{>=}", *op2]:
            return (
                constants.true
                if express(frame, op1).lower() >= bs_thing(frame, op2)[0].lower()
                else constants.false
            )
        case [op1, "caseless{<=}", *op2]:
            return (
                constants.true
                if express(frame, op1).lower() <= bs_thing(frame, op2)[0].lower()
                else constants.false
            )
        case [op1, "==", op2]:
            v1, v2 = express(frame, op1), express(frame, op2)
            return constants.true if v1 is v2 or v1 == v2 else constants.false
        case [op1, "!=", op2]:
            v1, v2 = express(frame, op1), express(frame, op2)
            return constants.true if v1 is v2 or v1 != v2 else constants.false
        case [op1, ">", op2]:
            return (
                constants.true
                if express(frame, op1) > express(frame, op2)
                else constants.false
            )
        case [op1, "<", op2]:
            return (
                constants.true
                if express(frame, op1) < express(frame, op2)
                else constants.false
            )
        case [op1, ">", "=", op2]:
            return (
                constants.true
                if express(frame, op1) >= express(frame, op2)
                else constants.false
            )
        case [op1, "<", "=", op2]:
            return (
                constants.true
                if express(frame, op1) <= express(frame, op2)
                else constants.false
            )
        case ["not", op1]:
            return constants.true if not express(frame, op1) else constants.false
        case [op1, "or", op2]:
            return (
                constants.true
                if express(frame, op1) or express(frame, op2)
                else constants.false
            )
        case [op1, "and", op2]:
            return (
                constants.true
                if express(frame, op1) and express(frame, op2)
                else constants.false
            )
        case [*op1, "in", op2]:
            return (
                constants.true
                if bs_thing(frame, op1)[0] in express(frame, op2)
                else constants.false
            )
        case ["LenOf", op1]:
            value = express(frame, op1)
            if hasattr(value, "__len__"):
                return len(value)
            else:
                return state.bstate("nil")
        case ["ToInt", op1]:
            value = express(frame, op1)
            try:
                return int(value)
            except:
                return state.bstate("nil")
        case ["ToFloat", op1]:
            value = express(frame, op1)
            try:
                return int(value)
            except:
                return state.bstate("nil")
        case ["ToStr", op1]:
            value = express(frame, op1)
            return str(value)
        case ["IsType", op1, t]:
            value = express(frame, op1)
            vtype = express(frame, t)
            return constants.true if isinstance(value, vtype) else constants.false
        case ["Append", lst, item]:
            (lst := expr_runtime(frame, lst)).append(express(frame, item))
            return lst
        case ["Pop", list(lst)]:
            return lst.pop() if lst else state.bstate("nil")
        case _:
            return state.bstate("nil")


def exprs_runtime(frame, args):
    "Process arguments at runtime"
    args = list(args)
    put = []
    res = []
    p = 0
    while p < len(args):
        c = args[p]
        if not isinstance(c, str):
            res.append(c)
        elif c.startswith("("):
            args[p] = c[1:]
            c = ""
            k = 1
            put.clear()
            while p < len(args) and k:
                c = args[p]
                put.append(str(c))
                p += 1
                if isinstance(c, str):
                    if c.startswith("("):
                        k += 1
                    elif c.endswith(")"):
                        k -= 1
            p -= 1
            put[-1] = put[-1][:-1]
            while "" in put:
                put.remove("")
            res.append(evaluate(frame, put))
        elif c.startswith("["):
            args[p] = c[1:]
            c = ""
            put.clear()
            k = 1
            while p < len(args) and k:
                c = args[p]
                put.append(str(c))
                p += 1
                if isinstance(c, str):
                    if c.startswith("["):
                        k += 1
                    elif c.endswith("]"):
                        k -= 1
            p -= 1
            put[-1] = put[-1][:-1]
            while "" in put:
                put.remove("")
            res.append([*bs_thing(frame, put)])
        elif c.startswith("!["):
            args[p] = c[2:]
            c = ""
            put.clear()
            k = 1
            while p < len(args) and k:
                c = args[p]
                put.append(str(c))
                p += 1
                if isinstance(c, str):
                    if c.startswith("["):
                        k += 1
                    elif c.endswith("]"):
                        k -= 1
            p -= 1
            put[-1] = put[-1][:-1]
            while "" in put:
                put.remove("")
            res.append((*bs_thing(frame, put),))
        elif c.startswith('"') and c.endswith('"'):
            text = c[1:-1]
            res.append(text)
        elif c.startswith("'") and c.endswith("'"):
            text = c[1:-1]
            d = flatten_dict(frame[-1]).items()
            for c, cc in d:
                text = text.replace(f"${{{c}}}", str(cc))
            res.append(text)
        elif c.startswith("<") and c.endswith(">"):
            res.append(c)
        else:
            res.append(expr_runtime(frame, c))
        p += 1
    return res


sep = " ,"
special_sep = "()+/-*[]<>?"


def group(text):
    res = []
    str_tmp = []
    id_tmp = []
    this = False
    rq = False
    quotes = {"str": '"', "pre": "}", "str1": "'"}
    str_type = "str"
    for i in text:
        if str_tmp:
            if this:
                if i == "n":
                    str_tmp.append("\n")
                elif i == "r":
                    str_tmp.append("\r")
                elif i == "t":
                    str_tmp.append("\t")
                elif i == "b":
                    str_tmp.append("\b")
                elif i == "f":
                    str_tmp.append("\f")
                elif i == "a":
                    str_tmp.append("\a")
                elif i == "v":
                    str_tmp.append("\v")
                else:
                    str_tmp.append(i)
                this = False
                continue
            if i == "\\":
                this = True
            elif i == quotes[str_type]:
                text = "".join(str_tmp) + quotes[str_type]
                if rq:
                    text = f'"{text}"'
                res.append(text)
                str_tmp.clear()
                rq = False
            else:
                str_tmp.append(i)
            continue
        elif i in sep:
            if id_tmp:
                res.append("".join(id_tmp))
                id_tmp.clear()
        elif i in special_sep:
            if id_tmp:
                res.append("".join(id_tmp))
                id_tmp.clear()
            res.append(i)
        elif i == "!":
            rq = True
        elif i in "\"{'":
            if id_tmp:
                res.append("".join(id_tmp))
                id_tmp.clear()
            str_tmp.append(i)
            if i == '"':
                str_type = "str"
            elif i == "{":
                str_type = "pre"
            elif i == "'":
                str_type = "str1"
            else:
                str_type = "str"
        else:
            id_tmp.append(i)
    if id_tmp:
        res.append("".join(id_tmp))
    return res


def exprs_preruntime(args):
    return [*map(expr_preruntime, args)]


def express(frame, e):
    return expr_runtime(frame, expr_preruntime(e))


def bs_thing(frame, e):
    return exprs_runtime(frame, exprs_preruntime(group(" ".join(map(str, e)))))



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension("math")

ext["memoize"] = {}


@ext.add_method(from_func=True, process=True)
@ext.add_func()
def expr(frame, __, expression):
    res = eval(expression, {"__builtins__": {}, **frame[-1]})
    if expression not in ext["memoize"]:
        ext["memoize"][expression] = res
    return (res,)



from os import name as os_name, system as cmd, path as os_path, get_terminal_size
from sys import argv


def batch_string(s, size):
    return [s[i : i + size] for i in range(0, len(s), size)]


local_path = os_path.join(os_path.dirname(__file__), "..", "lib", "docs")

(columns, rows) = get_terminal_size()


def clear_screen():
    if os_name == "nt":
        cmd("cls")
    else:
        cmd("clear")


# much more cleaner this way
def format_lines(current_data, lines):
    "Modify and format the lines in place"
    rt = []
    for line in lines:
        rt.append(expand(line))
    text = "\n".join(rt).strip().split("\n")
    del rt
    lines.clear()
    lines.extend(text)
    for pos, l in enumerate(lines):
        if l.startswith("^tag:"):
            current_data["tag:" + (name := l[5:].strip())] = pos
            lines[pos] = f"[ {name} ]".center(columns, "-")
        elif l.startswith("<tag:"):
            current_data["tag:" + (name := l[5:].strip())] = pos
            lines[pos] = f"--[ {name} ]".ljust(columns, "-")
        elif l.startswith(">tag:"):
            current_data["tag:" + (name := l[5:].strip())] = pos
            lines[pos] = f"[ {name} ]--".rjust(columns, "-")
        elif l.startswith("^txt:"):
            lines[pos] = l[5:].center(columns)
        elif l.startswith("<txt:"):
            lines[pos] = l[5:].ljust(columns)
        elif l.startswith(">txt:"):
            lines[pos] = l[5:].rjust(columns)
        elif l == ":border:":
            lines[pos] = "-" * columns


def expand(line):
    if len(line) <= columns:
        return line
    else:
        return "\n".join(batch_string(line, columns))


def format_code(lines):
    "Modify and format the lines in place"
    max_len = len(str(len(lines))) + 2
    lines.insert(
        0, "/" + ("-" * (max_len) + "+").ljust(columns - max_len + 1, "-") + "\\"
    )
    for pos, l in enumerate("\n".join(lines).strip().split("\n")[1:], 1):
        lines[pos] = f"| {pos} | {l}".ljust(columns - 1) + "|"
    lines.append("\\" + ("-" * (max_len) + "+").ljust(columns - max_len + 1, "-") + "/")


def page_print(d):
    text = (
        d["docu"]
        .replace("%page_title:-rjust", f' {d["title"]} --'.rjust(columns, "-"))
        .replace("%page_title:-ljust", f'-- {d["title"]} '.ljust(columns, "-"))
        .replace("%page_title:-center", f' {d["title"]} '.center(columns, "-"))
        .replace("%page_title:rjust", f' {d["title"]}   '.rjust(columns, "-"))
        .replace("%page_title:ljust", f'   {d["title"]} '.ljust(columns, "-"))
        .replace("%page_title:center", f' {d["title"]} '.center(columns, "-"))
        .replace("%page_title", d["title"])
        .replace("%page_author", d["author"])
    )
    clear_screen()
    if text.count("\n") + 2 < rows:
        print(text)
        input()
    else:
        text = text.split("\n")
        p = 0
        while True:
            clear_screen()
            print(*text[p : p + rows - 2], sep="\n")
            a = input()
            if a == "quit":
                clear_screen()
                return
            elif a == "home":
                p = 0
            elif a == "end":
                p = len(text) - 1
            elif a == "all":
                clear_screen()
                print(*text, sep="\n")
                return
            elif a.startswith("tag:"):
                if a in d:
                    p = d[a]
                else:
                    clear_screen()
                    input("Not found.\n")
            elif a.startswith(".") and a[1:] in d:
                clear_screen()
                input(d[a[1:]])
            elif p + 1 < len(text):
                p += 1


def mark_up(code):
    current_data = {
        "title": "default title",
        "author": "default_author",
        "gist": "default gist",  # short description
        "docu": "...",  # long description
        # Extra data can be entered
    }
    data = {}
    lines = []
    clines = []
    name = None
    ind = 0
    for pos, line in enumerate(code.split("\n")):
        if line.startswith("%rem%"):
            continue
        if clines:
            if line == "-end-":
                clines.pop(0)
                format_code(clines)
                lines.extend(clines)
                clines.clear()
            else:
                clines.append(line)
            continue
        elif line.startswith("::") and " " in line.strip():
            key, value = line[2:].split(maxsplit=1)
            data[key] = value
        elif line.startswith("::"):
            name = line[2:]
        elif line == "---" and name is not None:
            data[name] = "\n".join(lines).strip()
            lines.clear()
            name = None
        elif line.startswith(":") and " " in line:
            key, value = line[1:].split(maxsplit=1)
            current_data[key] = value
        elif line == "---":
            format_lines(current_data, lines)
            current_data["docu"] = "\n".join(lines).strip()
            lines.clear()
            data[current_data["title"].lower()] = current_data.copy()
            current_data.clear()
            current_data.update(
                {
                    "title": "default title",
                    "author": "default_author",
                    "gist": "default gist",  # short description
                    "docu": "...",  # long description
                    # Extra data can be entered
                }
            )
            ind = 0
        elif line == "-code-":
            clines.append("")
        elif line.strip() == "+ind":
            ind += 1
        elif line.strip() == "-ind" and ind > 0:
            ind -= 1
        elif line.strip() == "=ind" and ind > 0:
            ind = 0
        elif line.startswith("*"):
            lines.append("  " * (ind + 1) + line)
        else:
            lines.append("  " * ind + line)
    if lines:
        format_lines(current_data, lines)
        current_data["docu"] = "\n".join(lines).strip()
        data[current_data["title"].lower()] = current_data.copy()
    return data


def menu(file):
    data = mark_up(file)
    print("'quit' to stop.\n'list' to view items.")
    while True:
        act = input(": ").lower()
        if act == "quit":
            return
        elif act in data and isinstance(data[act], dict):
            page_print(data[act])
        elif act in data:
            print(data[act])
        elif act == "list":
            print(*map(str.title, data.keys()), sep="\n")
        else:
            print("Unknown action:", act)


def from_lib(path):
    file = os_path.join(local_path, path)
    if os_path.isfile(file):
        menu(open(file).read())
    else:
        print(f"{file!r} Does not exist.")


def from_local(file):
    if os_path.isfile(file):
        menu(open(file).read())
    else:
        print(f"{file!r} Does not exist.")



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension("array")


@ext.add_method(from_func=True)
@ext.add_func()
def slice(_, __, obj, from_, to=None):
    return (obj[from_:to],)


@ext.add_method(from_func=True)
@ext.add_func()
def join(_, __, obj, obj1):
    if not isinstance(obj1, (set, list, tuple)):
        obj1 = [obj1]
    if not isinstance(obj, (set, list, tuple)):
        obj = [obj]
    return (list(obj) + list(obj1),)


@ext.add_method(from_func=True)
@ext.add_func()
def reverse(_, __, obj):
    return (obj[::-1],)



#!/usr/bin/env python3

# DPL CLI
# We use match statements for te CLI
# To keep it lightweight, we dont need speed here.

import os
import sys

# DANGEROUS
sys.setrecursionlimit(10**6)
sys.set_int_max_str_digits(10**6)

import subprocess
import shutil
import time
import pstats
import lib.core.info as info
import lib.core.utils as utils
import lib.core.error as error
import lib.core.cli_arguments as cli_args
import lib.core.extension_support as ext_s
from dfpm import dfpm
from documenter import docs
import cProfile

try:  # Try to use the .pyd or .so parser to get some kick
    import lib.core.parser as parser
except Exception as e:  # fallback to normal python impl if it fails
    import lib.core.py_parser as parser
import lib.core.varproc as varproc
import lib.core.utils as utils

try:
    import dill as pickle

    has_dill = True
except ModuleNotFoundError:
    import pickle

    has_dill = False

ERRORS = {
    getattr(error, name): name
    for name in filter(lambda x: x.endswith("ERROR"), dir(error))
}


def rec(this, ind=0):
    if not isinstance(this, (tuple, list)):
        print(
            f"{'  '*ind}Error Name: {ERRORS.get(this, f'ERROR NAME NOT FOUND <{this}>')}"
        )
    else:
        for pos, i in enumerate(this):
            if isinstance(i, (tuple, list)):
                rec(i, ind + 1)
            else:
                print(
                    f"{'  '*ind}Error Name {'(original)' if pos == 0 else '(other)'}: {ERRORS.get(i, f'ERROR NAME NOT FOUND <{i}>')}"
                )


def ez_run(code, process=True, file="???", profile=False):
    if process:
        code = parser.process(code)
    if profile:
        stime = time.perf_counter()
    if err := parser.run(code):
        print(f"\n[{file}]\nFinished with an error: {err}")
        rec(err)
    if profile:
        delta = time.perf_counter() - stime
    parser.IS_STILL_RUNNING.set()
    parser.clean_threads()
    if profile:
        s, u = utils.convert_sec(delta)
        print(f"\nProgram Time: {s:,.2f}{u}")
    if err:
        exit(1)


def handle_args():
    if "arg-test" in varproc.flags:
        print("Flags:", varproc.flags)
        return
    if "info" in varproc.flags:
        info.print_info()
        return
    if "version" in varproc.flags or "v" in varproc.flags:
        print(
            f"DPL v{info.VERSION}\nUsing Python {info.PYTHON_VER}\n© Darren Chase Papa 2024\nMIT License (see LICENSE)"
        )
        return
    match (info.ARGV):
        case ["run", file, *args]:
            if not os.path.isfile(file):
                print("Invalid file path:", file)
                exit(1)
            if os.path.isfile("meta_config.cfg"):
                with open("meta_config.cfg", "r") as f:
                    varproc.meta = utils.parse_config(f.read(), {"meta": varproc.meta})[
                        "meta"
                    ]
            info.ARGV.clear()
            info.ARGV.extend([file, *args])
            varproc.meta["argc"] = info.ARGC = len(info.ARGV)
            with open(file, "r") as f:
                varproc.meta["internal"]["main_path"] = (
                    os.path.dirname(os.path.abspath(file)) + os.sep
                )
                varproc.meta["internal"]["main_file"] = file
                ez_run(
                    f.read(),
                    file=file,
                    profile="profile" in varproc.flags or "p" in varproc.flags,
                )
        case ["rc", file, *args]:
            if not os.path.isfile(file):
                print("Invalid file path:", file)
                exit(1)
            if os.path.isfile("meta_config.cfg"):
                with open("meta_config.cfg", "r") as f:
                    varproc.meta = utils.parse_config(f.read(), {"meta": varproc.meta})[
                        "meta"
                    ]
            info.ARGV.clear()
            info.ARGV.extend([file, *args])
            varproc.meta["argc"] = info.ARGC = len(info.ARGV)
            try:
                with open(file, "rb") as f:
                    code = pickle.loads(f.read())
                    varproc.meta["internal"]["main_file"] = file
                    varproc.meta["internal"]["main_path"] = (
                        os.path.dirname(os.path.abspath(file)) + os.sep
                    )
                    ez_run(
                        code,
                        False,
                        file,
                        profile="profile" in varproc.flags or "p" in varproc.flags,
                    )
            except Exception as e:
                print("Something went wrong:", file)
                print("Error:", repr(e))
                exit(1)
        case ["compile", file]:
            if not os.path.isfile(file):
                print("Invalid file path:", file)
                exit(1)
            output = file.rsplit(".", 1)[0] + ".cdpl"
            try:
                with open(file, "r") as in_file:
                    with open(output, "wb") as f:
                        f.write(pickle.dumps(parser.process(in_file.read())))
            except Exception as e:
                print("Something went wrong:", file)
                print("Error:", repr(e))
                exit(1)
        case ["docs", name]:
            docs.from_lib(name)
        case ["docu", name]:
            docs.from_local(name)
        case ["build", python_bin]:
            print(
                f"This will build a compiled parser for your system.\nThis does not necessarily mean that the parser will be faster!\nThis will build for {python_bin}\nCython does not like loading functions from exec, please be careful."
            )
            if input("Proceed? [y/N] ").strip().lower() in {"y", "yes"}:
                import lib.core.build as bld

                bld.run(python_bin)
        case ["build-clean"]:
            print(
                "Removing any files generated by cython (including the built parser) in the `./lib/core` directory"
            )
            stuff = set()
            for i in os.listdir(info.CORE_DIR):
                stuff.add(os.path.join(os.getcwd(), info.CORE_DIR, i))
            if os.path.isdir(temp := os.path.join(info.CORE_DIR, "build")):
                print("Removed build directory made by cython.")
                shutil.rmtree(temp)
            stuff = (
                *filter(
                    lambda x: x.rsplit(".", 1)[-1] in {"so", "pyd", "c", "pyi", "dll"},
                    stuff,
                ),
            )
            for pos, i in enumerate(stuff):
                print(f"[{((pos+1)/len(stuff))*100:,.2f}] Removing:", i)
                try:
                    os.remove(i)
                except:
                    print("Failed. Another process may be using it! Terminate it.")
        case ["install", python_exec]:
            print(f"Installing requirements for `{sys.platform}`")
            tmp_dir = os.getcwd()
            if info.BINDIR:
                os.chdir(info.BINDIR)
            with open("requirements.txt", "r") as f:
                while (line := f.readline().strip()) != "end":
                    if "#" in line:
                        line = line[: line.index("#")].strip()
                    if not line:
                        continue
                    if line.startswith("?"):
                        if "verbose" in varproc.flags:
                            print(
                                f"Conditional install [for {(temp:=line[1:line.index(' ')])}{' (match)' if temp == sys.platform or temp == 'any' else ' (mismatch)'}]: {line[len(sys.platform)+2:]}"
                            )
                        if (
                            line[1:].startswith(sys.platform)
                            or line[1 : line.index(" ")] == "any"
                        ):
                            line = line[len(sys.platform) + 1 :].strip()
                        else:
                            continue
                    elif line.startswith("!"):
                        if "verbose" in varproc.flags:
                            print(
                                f"Conditional command [for {(temp:=line[1:line.index(' ')])}{' (match)' if temp == sys.platform or temp == 'any' else ' (mismatch)'}]: {line[len(sys.platform)+2:]}"
                            )
                        if (
                            line[1:].startswith(sys.platform)
                            or line[1 : line.index(" ")] == "any"
                        ):
                            line = line[len(sys.platform) + 2 :].strip()
                            if "verbose" in varproc.flags:
                                print(f"Running: {line}")
                            if err := os.system(line):
                                print(f"Error code: {err}")
                            continue
                    with open(os.devnull, "w") as devnull:
                        print("Installing:", line)
                        try:
                            if subprocess.run(
                                [
                                    python_exec,
                                    "-m",
                                    "pip",
                                    "install",
                                    "--ignore-installed",
                                    line,
                                ],
                                stdout=devnull,
                                stderr=devnull,
                            ).returncode:
                                print(f"Error while installing: {line}")
                        except Exception as e:
                            print(
                                f"Failed to install [{line}]: {repr(e)}\nPlease check the usage of `dpl.py install`"
                            )
            print("Done!")
            os.chdir(tmp_dir)
        case ["package", *args]:
            match args:
                case ["install", user, repo]:
                    dfpm.dl_repo(user, repo, location=info.LIBDIR)
                case ["install", user, repo, branch]:
                    dfpm.dl_repo(user, repo, branch, location=info.LIBDIR)
                case ["installto:", ipath, user, repo]:
                    dfpm.dl_repo(user, repo, location=ipath)
                case ["installto:", ipath, user, repo, branch]:
                    dfpm.dl_repo(user, repo, branch, location=ipath)
                case ["install", user, repo, branch, use]:
                    dfpm.dl_repo(
                        user,
                        repo,
                        branch,
                        location=info.LIBDIR,
                        use_branch_name=use == "true",
                    )
                case ["installto:", ipath, user, repo, branch, use]:
                    dfpm.dl_repo(
                        user,
                        repo,
                        branch,
                        location=ipath,
                        use_branch_name=use == "true",
                    )
                case ["remove", pack_name]:
                    if not os.path.isdir(
                        pack_path := os.path.join(info.LIBDIR, pack_name)
                    ):
                        print("Package doesnt exist!")
                        return
                    print(pack_path, "Is going to be removed.")
                    if input("Enter y to continue: ").lower() in {"y", "yes"}:
                        dfpm.delete(pack_path)
                    print("Done!")
                case _:
                    print("Invalid command!")
                    return
        case ["repr"] | []:
            if os.path.isfile(os.path.join(info.BINDIR, "start_prompt.txt")):
                start_text = open(os.path.join(info.BINDIR, "start_prompt.txt")).read()
            else:
                start_text = ""
            frame = varproc.new_frame()
            if "import-all" in varproc.flags:
                if "verbose" in varproc.flags:
                    print("Importing all standard modules...")
                for pos, file in enumerate(
                    temp := varproc.meta["internal"]["libs"]["std_libs"], 1
                ):
                    if "verbose" in varproc.flags:
                        print(f"[{(pos/len(temp))*100:7.2f}% ] Importing: {file}")
                    ext_s.py_import(frame, file, "@std")
                if "verbose" in varproc.flags:
                    print("Done importing!")
            PROMPT_CTL = frame[-1]["_meta"]["internal"]["prompt_ctl"] = {}
            PROMPT_CTL["ps1"] = ">>> "
            PROMPT_CTL["ps2"] = "... "
            print(
                f"DPL REPL for DPL {varproc.meta['internal']['version']}\nPython {info.PYTHON_VER}{(chr(10)+start_text) if start_text else ''}"
            )
            START_FILE = os.path.join(info.BINDIR, "start_script.dpl")
            if os.path.isfile(START_FILE):
                try:
                    with open(START_FILE, "r") as f:
                        parser.run(parser.process(f.read(), name="dpl_repl-startup"))
                except:
                    print("something went wrong while running start up script!")
            while True:
                try:
                    act = input(PROMPT_CTL["ps1"]).strip()
                except KeyboardInterrupt:
                    exit()
                if (
                    act
                    and (
                        (temp := act.split(maxsplit=1)[0]) in info.INC
                        or temp in info.INC_EXT
                    )
                    or act == "#multiline"
                ):
                    while True:
                        try:
                            aa = input(PROMPT_CTL["ps2"])
                        except KeyboardInterrupt:
                            exit()
                        if not aa:
                            break
                        act += "\n" + aa
                elif act == ".paste":
                    act = ""
                    while True:
                        act += (this := input())
                        if not this:
                            break
                elif act == "exit":
                    break
                elif act.startswith("$"):
                    try:
                        err = os.system(act[1:])
                    except BaseException as e:
                        err = f"Error Raised: {repr(e)}"
                    finally:
                        print("\nDone!")
                    if err:
                        print(f"Error Code: {err}")
                    else:
                        print("Success")
                    continue
                elif act == ".reload":
                    if os.path.isfile(START_FILE):
                        try:
                            with open(START_FILE, "r") as f:
                                parser.run(
                                    parser.process(f.read(), name="dpl_repl-startup")
                                )
                        except:
                            print("something went wrong while running start up script!")
                    continue
                try:
                    if err := parser.run(parser.process(act), frame=frame):
                        rec(err)
                except Exception as e:
                    print(f"Python Exception was raised while running:\n{repr(e)}")
        case ["help"]:
            print(
                f"""Help for DPL [v{varproc.meta['internal']['version']}]

dpl run [file] args...
    Runs the given DPL script.
dpl rc [file] args...
    Runs the given compiled DPL script.
dpl compile [file]
    Compiles the given DPL script.
    Outputs to [file].cdpl
dpl install [python_exec]
    Installs runtime requirements.
    See the requirements.txt file.
dpl build [python executable]
    Builds the parser and cythonizes it.
    The interpreter chooses which to run automatically.
    Although it might be changeable in the configs soon!
dpl build-clean
    Removes the cythonized components.
dpl repr ALSO JUST `dpl`
    Invokes the REPL
dpl package install <user> <repo> <branch> <include_branch_name?>
    Install a package hosted on github.
    Default branch is 'master'
dpl package installto: <path_to_dest> <user> <repo> <branch> <include_branch_name?>
    Install a package hosted on github.
    Default branch is 'master'
dpl package remove <package_name>
    Delete that package.
dpl docs doc_name.mmu
dpl docu doc_name.mmu
dpl -info
    Prints info.
dpl -arg-test
    Tests flag handling.
'dpl -version' or 'dpl -v'
    Prints version and some info.
'dpl -profile ...' or 'dpl -p ...'
    Profiles the code using 'time.perf_counter' for inaccurate but fast execution.
dpl -cprofile ...
    Profiles the code using cProfile for more accurate but slower execution.
"""
            )
        case _:
            print("Invalid invokation!")
            print("See 'dpl help' for more")
            exit(1)
    if "pause" in varproc.flags:
        input("\n[Press Enter To Finish]")


if __name__ == "__main__":
    flags = cli_args.flags(info.ARGV, True)
    varproc.flags.update(flags)
    info.ARGC = len(info.ARGV)
    if "cprofile" in flags:
        profiler = cProfile.Profile()
        profiler.enable()
    handle_args()
    if "cprofile" in flags:
        profiler.disable()
        default = "tottime"
        order_by = None
        for i in flags:
            if i.startswith("order_profile="):
                order_by = i[14:]
        print("\nProfile Result")
        stats = pstats.Stats(profiler)
        stats.sort_stats(order_by or default).print_stats()



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension("string")


@ext.add_method(from_func=True)
@ext.add_func()
def split(_, __, string, delim=" ", times=-1):
    return (string.split(delim, times),)


@ext.add_method(from_func=True)
@ext.add_func()
def lower(_, __, string):
    return (string.lower(),)


@ext.add_method(from_func=True)
@ext.add_func()
def upper(_, __, string):
    return (string.upper(),)


@ext.add_method(from_func=True)
@ext.add_func()
def isLower():
    return (string.islower(),)


@ext.add_method(from_func=True)
@ext.add_func()
def isUpper():
    return (string.isupper(),)


@ext.add_method(from_func=True)
@ext.add_func()
def clean(_, __, string):
    return (string.strip(),)


@ext.add_func("myHash")
def myHash(_, __, string):
    "THIS IS NOT A PERFECT HASH FUNCTION AND JUST USES PYTHONS HASH FUNC\nTHIS WILL BE REPLACED WITH sha256 SOON"
    return (hash(string),)



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension(meta_name="io")


@ext.add_func("open")
def myOpen(_, local, file_name, mode="r"):
    try:
        if modules.os.path.isabs(file_name):
            file = file_name
        else:
            file = modules.os.path.join(local, file_name)
        return (open(file),)
    except Exception as e:
        return (e,)


@ext.add_func()
def seek(_, __, file_object, position, whence=0):
    file_object.seek(position, whence)


@ext.add_func()
def read(_, __, file_object):
    try:
        return (file_object.read(),)
    except Exception as e:
        file_object.close()
        return (e,)


@ext.add_func()
def write(_, __, file_object, content):
    try:
        file_object.write(content)
    except Exception as e:
        file_object.close()
        return f"err:{dpl.error.PYTHON_ERROR}:{repr(e)}"


@ext.add_func()
def append(_, __, file_object, content):
    try:
        if (mode := file_object.mode) == "a":
            file_object.append(content)
        else:
            raise Exception(
                f'Invalid operation on a file! Expected the mode to be "a" but got "{mode}"'
            )
    except Exception as e:
        file_object.close()
        return f"err:{dpl.error.PYTHON_ERROR}:{repr(e)}"


@ext.add_func()
def close(_, __, file_object):
    file_object.close()



# Utilities for QOL functions
# Now includes DSL Config Parser

from . import arguments as argproc

try:
    import hashlib

    has_hash = True
except ImportError:
    hash_hash = False


def flatten_dict(d, parent_key="", sep="."):
    items = {}
    for key, value in d.items():
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(flatten_dict(value, new_key, sep=sep))
        else:
            items[new_key] = value
    return items


def get_val(dct, full_name, default=0, sep="."):
    "Get a variable"
    if "." not in full_name:
        return dct.get(full_name, default)
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if pos != last and name in node and isinstance(node[name], dict):
            node = node[name]
        elif pos == last and name in node:
            return node[name]
        else:
            return default
    return default


def set_val(dct, full_name, value, sep="."):
    "Set a variable"
    if "." not in full_name:
        dct[full_name] = value
        return
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if pos != last and name in node and isinstance(node[name], dict):
            node = node[name]
        elif pos == last:
            node[name] = value


def constant_hash(obj):
    "Constant hash function"
    assert has_hash, "hashlib is unavailable."
    obj_representation = repr(obj).encode("utf-8")
    return hashlib.sha256(obj_representation).hexdigest()


def convert_sec(sec):
    "Convert seconds to appropriate units"
    if sec >= 1:
        return sec, "s"
    elif sec >= 1e-3:
        return sec * 1e3, "ms"
    elif sec >= 1e-6:
        return sec * 1e6, "µs"
    elif sec >= 1e-9:
        return sec * 1e9, "ns"
    else:
        return sec * 1e12, "ps"


def convert_bytes(byte):
    "Convert bytes to appropriate units"
    if byte < 1e3:
        return byte, "B"
    elif byte < 1e6:
        return byte * 1e-3, "KB"
    elif byte < 1e9:
        return byte * 1e-6, "MB"
    elif byte < 1e12:
        return byte * 1e-9, "GB"
    elif byte < 1e15:
        return byte * 1e-12, "TB"
    else:
        return byte * 1e-15, "PB"


def parse_config(code, format=None):
    name = None
    lines = code.split("\n")
    data = {} if format is None else format
    for line in lines:
        line = line.lstrip()
        if not line or line.startswith("#"):
            continue
        elif line == "end" and name is not None:
            name = None
        elif line.startswith("[") and line.endswith("]"):
            name = line[1:-1]
            if get_val(data, name, default=None) is None:
                set_val(data, name, {})
        else:
            ins, op, value = line.split(maxsplit=2)
            if argproc.is_int(value):
                value = int(value)
            elif argproc.is_float(value):
                value = float(value)
            elif value == "@list":
                value = []
            elif value == "@dict":
                value = {}
            elif value.startswith('"') and value.endswith('"'):
                value = value[1:-1]
            else:
                value = ""
            if op == "=":
                set_val(data, (name + "." if name is not None else "") + ins, value)
            elif op == "append":
                temp = get_val(data, (name + "." if name is not None else "") + ins)
                if temp is not None:
                    temp.append(value)
                else:
                    set_val(
                        data, (name + "." if name is not None else "") + ins, [value]
                    )
            else:
                pass
    return data



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

type_vars = dpl.extension()


def check(var, value):
    try:
        return True if isinstance(value, var["type"]) else False
    except:
        raise Exception("Type mismatch!")


@type_vars.add_func()
def defv(frame, _, name, value_type):
    dpl.varproc.rset(
        frame[-1], name, {"type": value_type, "[meta_value]": value_type()}
    )


@type_vars.add_func()
def setv(frame, _, name, value):
    old = dpl.varproc.rget(frame[-1], name, meta=False)
    if old == dpl.state_nil:
        return f"err:{dpl.error.NAME_ERROR}:Variable {name!r} is not defined!"
    else:
        if not check(old, value):
            return f"err:{dpl.error.TYPE_ERROR}:Variable {name!r} expects type {old['type']} but got {type(value)} ({value!r})"
        dpl.varproc.rset(frame[-1], name, value)



# A module to help the parser call dynamically loaded functions and to load them too
# This is needed for the cythonized parser and slows down the pure python impl!
# I tried it, it still failed :)

from . import utils
from . import varproc
from . import arguments as argproc
from . import info
from . import error
from . import state
from . import restricted
from . import objects
import types
import itertools
import time
import os, sys
import traceback
import __main__


def register_run(func):
    dpl.run_code = func


def register_process(func):
    dpl.process_code = func


class modules:
    "Capsule for modules."

    os = os
    sys = sys
    traceback = traceback
    time = time
    types = types
    itertools = itertools


class extension:
    "A class to help define methods and functions."

    def __init__(self, name=None, meta_name=None):
        self.__func = {}  # functions
        self.__meth = {}  # methods
        self.name = (
            name  # This is a scope name,              dpl defined name.func_name
        )
        self.meta_name = meta_name  # while this is the mangled name, python defined "{meta_name}:{func_name}"

    def add_func(self, name=None):
        "Add a function."

        def wrap(func):
            nonlocal name
            if func.__doc__ is None:
                func.__doc__ = (
                    f"Function {self.meta_name}:{name}"
                    if self.meta_name
                    else f"{self.name}.{name}"
                ) + ": Default doc string..."
            if name is None:
                name = getattr(func, "__name__", None) or "_"
            self.__func[name if not self.meta_name else f"{self.meta_name}:{name}"] = (
                func
            )
            return func

        return wrap

    def add_method(self, name=None, process=True, from_func=False):
        "Add a method."

        def wrap(func):
            nonlocal name
            if name is None:
                name = getattr(func, "__name__", None) or "_"
            self.__meth[
                (
                    f"{self.name}.{name}"
                    if not self.meta_name
                    else f"{self.meta_name}:{name}"
                )
            ] = process, (
                func
                if not from_func
                else lambda *args: func(args[0], None, *args[1:])[0]
            )
            return func

        return wrap

    def __setitem__(self, name, value):
        self.__func[name if not self.meta_name else f"{self.meta_name}:{name}"] = value

    def __getitem__(self, name):
        return self.__func[name if not self.meta_name else f"{self.meta_name}:{name}"]

    def get(self, name, default=None):
        return self.__data.get(name, default)

    @property
    def functions(self):
        return self.__func

    @property
    def methods(self):
        return self.__meth


def require(path):
    "Import a python in the lib dir.\nIn cases of 'dir/.../file' use ['dir', ..., 'file'],\nthis uses os.path.join to increase portability."
    mod = {
        "__name__": "__dpl_require__",
        "modules": modules,
        "dpl": dpl,
        "__import__": restricted.restricted(__import__),
    }
    if isinstance(path, (list, tuple)):
        path = os.path.join(*path)
    try:
        with open(os.path.join(info.LIBDIR, path), "r") as f:
            exec(compile(f.read(), path, "exec"), mod)
        r = types.ModuleType(path)
        for name, value in mod.items():
            setattr(r, name, value)
        return r
    except:
        return None


class dpl:
    require = require
    utils = utils
    varproc = varproc
    arguments = argproc
    info = info
    error = error
    state = state
    restricted = restricted
    state_nil = state.bstate("nil")
    state_none = state.bstate("none")
    state_true = 1
    state_false = 0
    extension = extension
    objects = objects
    falsy = (state_nil, state_none, state_false, None, False)
    truthy = (state_true, True)


def py_import(frame, file, search_path=None, loc=varproc.meta["internal"]["main_path"]):
    if not os.path.isabs(file):
        if search_path is not None:
            file = os.path.join(
                {"_std": varproc.meta["internal"]["lib_path"], "_loc": loc}.get(
                    search_path, search_path
                ),
                file,
            )
        if not os.path.isfile(file):
            print("File not found:", file)
            return 1
    if varproc.is_debug_enabled("show_imports"):
        error.info(f"Imported {file!r}")
    with open(file, "r") as f:
        obj = compile(f.read(), file, "exec")
        try:
            d = {"__name__": "__dpl__", "modules": modules, "dpl": dpl}
            exec(obj, d)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            error.error("[N/A]", file, traceback.format_exc())
            return 1
    funcs = {}
    meths = {}
    for name, ext in d.items():
        if isinstance(ext, extension):
            if ext.name in frame[-1]:
                raise Exception(f"Name clashing! For name {ext.name!r}")
            if ext.name:
                varproc.rset(frame[-1], ext.name, (temp := {}))
                temp.update(ext.functions)
            else:
                funcs.update(ext.functions)
            meths.update(ext.methods)
    frame[-1].update(funcs)
    argproc.methods.update(meths)
    file = os.path.realpath(file)
    if search_path in varproc.meta["dependencies"]["python"]:
        varproc.dependencies["python"][search_path].add(file)
    else:
        varproc.dependencies["python"][search_path] = {file}


def py_import_string(
    frame, file_name, code, search_path=None, loc=varproc.meta["internal"]["main_path"]
):
    if not os.path.isabs(file_name):
        if search_path is not None:
            file = os.path.join(
                {"@lib": varproc.meta["internal"]["lib_path"], "@loc": loc}.get(
                    search_path, search_path
                ),
                file_file,
            )
        if not os.path.isfile(file_name):
            print("File not found:", file_file)
            return 1
    if varproc.is_debug_enabled("show_imports"):
        error.info(f"Imported {file_name!r}")
    obj = compile(code, file_name, "exec")
    try:
        d = {"modules": modules, "dpl": dpl}
        d["__name__"] = "__dpl__"
        exec(obj, d)
    except (SystemExit, KeyboardInterrupt):
        raise
    except:
        error.error("[N/A]", file_name, traceback.format_exc())
        return 1
    funcs = {}
    meths = {}
    for name, ext in d.items():
        if isinstance(ext, extension):
            if ext.name in frame[-1]:
                raise Exception(f"Name clashing! For name {ext.name!r}")
            funcs.update(ext.functions)
            meths.update(ext.methods)
    frame[-1].update(funcs)
    argproc.methods.update(meths)


def call(func, frame, file, args, kwargs={}):
    if varproc.is_debug_enabled("track_time"):
        start = time.time()
    ret = func(frame, file, *args, **kwargs)
    if varproc.is_debug_enabled("track_time"):
        delta = time.time() - start
        if delta > varproc.get_debug("time_threshold"):
            delta_value, delta_unit = utils.convert_sec(delta)
            error.info(
                f"The function {func} took too long!\nPrecisely: {delta_value:,.8f}{delta_unit}"
            )
    return ret


def call_w_body(func, frame, file, body, args, kwargs={}):
    if varproc.is_debug_enabled("track_time"):
        start = time.time()
    ret = func(frame, file, body, *args, **kwargs)
    if varproc.is_debug_enabled("track_time"):
        delta = time.time() - start
        if delta > varproc.get_debug("time_threshold"):
            delta_value, delta_unit = utils.convert_sec(delta)
            error.info(
                f"The function {func} took too long!\nPrecisely: {delta_value:,.8f}{delta_unit}"
            )
    return ret



from setuptools import setup

setup(
    name="duprol",
    version="1.4.5",
    install_requires=["dill"],  # Add dependencies if needed
    author="Darren Chase Papa",
    author_email="darrenchasepapa@gmail.com",
    description="A dumb programming language",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



# Logging

import datetime
import os
from . import varproc

ERRORS = (
    "PREPROCESSING_ERROR",
    "SYNTAX_ERROR",
    "RUNTIME_ERROR",
    "PYTHON_ERROR",
    "PANIC_ERROR",
    "ASSERT_ERROR",
    "IMPORT_ERROR",
    "THREAD_ERROR",
    "TYPE_ERROR",
    "NAME_ERROR",
    "COMPAT_ERROR",
)


META_ERR = varproc.meta["err"] = {"defined_errors": ERRORS}

for pos, name in enumerate(ERRORS, 1):
    globals()[name] = pos
    META_ERR[name] = pos

STOP_RESULT = -1
SKIP_RESULT = -2

ERRORS_DICT = {
    globals().get(name): name for name in filter(lambda x: x.endswith("ERROR"), dir())
}

output = varproc.get_debug("debug_output_file")

if os.path.isfile(output):
    os.remove(output)


def my_print(text):
    if varproc.is_debug_enabled("log_events") and not os.path.isfile(output):
        with open(output, "w") as f:
            f.write(
                f"""[LOG FILE]
AUTOMATICALLY GENERATED LOG FILES BY THE INTERPRETER!
ANY LOG OUTPUT WAS REDIRECTED TO HERE!
Log file path: {output}
Date of initial logs: {datetime.datetime.now()}
Debug options:
"""
            )
        with open(output, "a") as f:
            for name, enabled in varproc.debug.items():
                if isinstance(enabled, str):
                    f.write(f"  value: {enabled!r} - {name!r}\n")
                elif isinstance(enabled, int):
                    f.write(f"  {'on ' if enabled else 'off'} - {name!r}\n")
    if varproc.is_debug_enabled("log_events"):
        with open(output, "a") as f:
            f.write(f"{text}\n")
    else:
        print(text)


# pre_[name] is a preprocessing log


def pre_error(pos, file, cause=None):
    my_print(f"\n[Preprocessing Error]\nError in line {pos} file {file!r}")
    if cause is not None:
        my_print(f"Cause:\n{cause}")


def error(pos, file, cause=None):
    my_print(f"\nError in line {pos} file {file!r}")
    if cause is not None:
        my_print(f"Cause:\n{cause}")


def info(text, show_date=True):
    if show_date:
        my_print(f"   [INFO] {datetime.datetime.now()}: {text}")
    else:
        my_print(f"   [INFO]: {text}")


def warn(text, show_date=True):
    if show_date:
        my_print(f"[WARNING] {datetime.datetime.now()}: {text}")
    else:
        my_print(f"[WARNING]: {text}")


def pre_info(text, show_date=True):
    if show_date:
        my_print(f"   [INFO PRE] {datetime.datetime.now()}: {text}")
    else:
        my_print(f"   [INFO PRE]: {text}")


def pre_warn(text, show_date=True):
    if show_date:
        my_print(f"[WARNING PRE] {datetime.datetime.now()}: {text}")
    else:
        my_print(f"[WARNING PRE]: {text}")



if __name__ != "__dpl_require__":
    raise Exception("This must be included by a DuProL Extension script!")


def has_repr(obj):
    return "_internal" in obj and "_im_repr" in obj


def get_repr(func):
    frame = dpl.varproc.new_frame()
    frame[0]["_returns"] = ["result"]
    if func["self"] != dpl.state.bstate("nil"):
        dpl.varproc.rset(frame[-1], "self", func["self"])
    err = dpl.run_code(func["body"], frame)
    if err:
        raise Exception(err)
    return dpl.varproc.rget(frame[0], "result")



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

helper = dpl.require(["dpl_helpers", "func_helper.py"])

if helper is None:
    raise Exception("Helper func_helper.py doesnt exist!")

ext = dpl.extension(meta_name="io")

ext["output"] = modules.sys.stdout


@ext.add_func("print")
def myPrint(_, __, *args, end="", sep=" "):
    args = list(args)
    for pos, arg in enumerate(args):
        if isinstance(arg, dict) and helper.has_repr(arg):
            arg[pos] = helper.get_repr(arg["_im_repr"])
    print(*args, end=end, sep=sep, file=ext["output"])


@ext.add_func()
def printf(_, __, text, values):
    for name, value in values.items():
        text = text.replace(f"${{{name}}}", str(value))
    print(text, end="", file=ext["output"])


@ext.add_func()
def println(_, __, *args, sep=" "):
    args = list(args)
    for pos, arg in enumerate(args):
        if isinstance(arg, dict) and helper.has_repr(arg):
            args[pos] = helper.get_repr(arg["_im_repr"])
    print(*args, sep=sep, file=ext["output"])


@ext.add_func()
def rawprint(_, __, *args, sep=" ", end=""):
    print(*args, sep=sep, file=ext["output"], end=end)


@ext.add_func()
def rawprintln(_, __, *args, sep=" "):
    print(*args, sep=sep, file=ext["output"])


@ext.add_func("input")
def myInput(frame, __, prompt=None, name=None):
    res = input(prompt if prompt not in dpl.falsy else "")
    if name is not None:
        dpl.varproc.rset(frame[-1], name, res)


@ext.add_func()
def test(_, __, test):
    print(helper.has_repr(test))


@ext.add_func()
def setOutputFile(_, __, file):
    ext["output"] = file


@ext.add_func()
def rawoutput(_, __, *values):
    s = []
    for i in values:
        if isinstance(i, int):
            try:
                s.append(chr(i))
            except:
                s.append(str(i))
        else:
            s.append(str(i))
    modules.sys.stdout.write("".join(s))
    modules.sys.stdout.flush()


@ext.add_func()
def flush(_, __):
    modules.sys.stdout.flush()


# misc


@ext.add_func()
def printf(_, __, fmt, *values, sep=" ", end=""):
    for pos, value in enumerate(values):
        fmt = fmt.replace(f"%{pos}", str(value))
    print(fmt, sep=sep, end=end)
